/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2010, Digium, Inc.
 *
 * Marco Signorini <marco.signorini@loway.ch>
 * based on 
 * David Vossel <dvossel@digium.com>
 * func_frame_trace.c
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Triggers dialplan actions when specific CONTROL_FRAMES are detected on a channel.
 *
 * \author Marco Signorini <marco.signorini@loway.ch>
 *
 * \ingroup functions
 */

/*** MODULEINFO
	<support_level>extended</support_level>
 ***/

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 336316 $")

#include "asterisk/module.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/framehook.h"
#include "asterisk/stringfields.h"
#include "asterisk/app.h"


/*** DOCUMENTATION
	<function name="FRAME_TRIGGER" language="en_US">
		<synopsis>
			Triggers dialplan actions when specific CONTROL_FRAMES are detected on a channel.
		</synopsis>
		<syntax>
			<parameter name="filter list type" required="true">
				<para>A filter can be applied to the trigger to limit frames sensibility.  This
				filter can either be a <literal>white</literal> or <literal>black</literal> list
				of control frame types.  When no filter type is present, <literal>white</literal> is
				used.  If no arguments are provided at all, all frames will trigger the specified gosub call.
                If the specified gosub extension is not present on the dialplan, the function will gosub to the
                "s-XXXX" extension in the gosub context.
				</para>

				<para>Below are the different types of control frames that can be filtered.</para>
				<enumlist>
					<enum name = "CONTROL_HANGUP" />
					<enum name = "CONTROL_RING" />
					<enum name = "CONTROL_RINGING" />
					<enum name = "CONTROL_ANSWER" />
					<enum name = "CONTROL_BUSY" />
					<enum name = "CONTROL_TAKEOFFHOOK" />
					<enum name = "CONTROL_OFFHOOK" />
					<enum name = "CONTROL_CONGESTION" />
					<enum name = "CONTROL_FLASH" />
					<enum name = "CONTROL_WINK" />
					<enum name = "CONTROL_OPTION" />
					<enum name = "CONTROL_RADIO_KEY" />
					<enum name = "CONTROL_RADIO_UNKEY" />
					<enum name = "CONTROL_PROGRESS" />
					<enum name = "CONTROL_PROCEEDING" />
					<enum name = "CONTROL_HOLD" />
					<enum name = "CONTROL_UNHOLD" />
					<enum name = "CONTROL_VIDUPDATE" />
					<enum name = _XXX_"CONTROL_T38" />
					<enum name = "CONTROL_SRCUPDATE" />
					<enum name = "CONTROL_TRANSFER" />
					<enum name = "CONTROL_CONNECTED_LINE" />
					<enum name = "CONTROL_REDIRECTING" />
					<enum name = "CONTROL_T38_PARAMETERS" />
					<enum name = "CONTROL_CC" />
					<enum name = "CONTROL_SRCCHANGE" />
					<enum name = "CONTROL_READ_ACTION" />
					<enum name = "CONTROL_AOC" />
					<enum name = "CONTROL_END_OF_Q" />
					<enum name = "CONTROL_INCOMPLETE" />
					<enum name = "CONTROL_MCID" />
					<enum name = "CONTROL_UPDATE_RTP_PEER" />
				</enumlist>
			</parameter>
		</syntax>
		<description>
			<para>Examples:</para>
			<para>exten => 1,1,Set(FRAME_TRIGGER(white, CONTROL_HOLD)=ext,context); trigger gosub ext-CONTROL_HOLD@context only for HOLD type control frames </para>
			<para>exten => 1,1,Set(FRAME_TRIGGER(black, CONTROL_HOLD)=ext,context); trigger gosub ext-CONROL_XXXX@context for all control type frames except for HOLD type control frames </para>
			<para>exten => 1,1,Set(FRAME_TRIGGER(CONTROL_HOLD, CONROL_UNHOLD)=context) ; trigger gosub to s-CONTROL_HOLD@context and s-CONROL_UNHOLDcontext only for HOLD and UNHOLD type control frames </para>
		</description>
	</function>
 ***/

static struct {
	enum ast_control_frame_type type;
	const char *str;
} controltype2str[] = {
	{ AST_CONTROL_HANGUP, "CONTROL_HANGUP" },
	{ AST_CONTROL_RING, "CONTROL_RING" },
	{ AST_CONTROL_RINGING, "CONTROL_RINGING" },
	{ AST_CONTROL_ANSWER, "CONTROL_ANSWER" },
	{ AST_CONTROL_BUSY, "CONTROL_BUSY" },
	{ AST_CONTROL_TAKEOFFHOOK, "CONTROL_TAKEOFFHOOK" },
	{ AST_CONTROL_OFFHOOK, "CONTROL_OFFHOOK" },
	{ AST_CONTROL_CONGESTION, "CONTROL_CONGESTION" },
	{ AST_CONTROL_FLASH, "CONTROL_FLASH" },
	{ AST_CONTROL_WINK, "CONTROL_WINK" },
	{ AST_CONTROL_OPTION, "CONTROL_OPTION" },
	{ AST_CONTROL_RADIO_KEY, "CONTROL_RADIO_KEY" },
	{ AST_CONTROL_RADIO_UNKEY, "CONTROL_RADIO_UNKEY" },
	{ AST_CONTROL_PROGRESS, "CONTROL_PROGRESS" },
	{ AST_CONTROL_PROCEEDING, "CONTROL_PROCEEDING" },
	{ AST_CONTROL_HOLD, "CONTROL_HOLD" },
	{ AST_CONTROL_UNHOLD, "CONTROL_UNHOLD" },
	{ AST_CONTROL_VIDUPDATE, "CONTROL_VIDUPDATE" },
	{ _XXX_AST_CONTROL_T38, "AST_CONTROL_T38" },
	{ AST_CONTROL_SRCUPDATE, "CONTROL_SRCUPDATE" },
	{ AST_CONTROL_TRANSFER, "CONTROL_TRANSFER" },
	{ AST_CONTROL_CONNECTED_LINE, "CONTROL_CONNECTED_LINE" },
	{ AST_CONTROL_REDIRECTING, "CONTROL_REDIRECTING" },
	{ AST_CONTROL_T38_PARAMETERS, "CONTROL_T38_PARAMETERS" },
	{ AST_CONTROL_CC, "CONTROL_CC" },
	{ AST_CONTROL_SRCCHANGE, "CONTROL_SRCCHANGE" },
	{ AST_CONTROL_READ_ACTION, "CONTROL_READ_ACTION" },
	{ AST_CONTROL_AOC, "CONTROL_AOC" },
	{ AST_CONTROL_END_OF_Q, "CONTROL_END_OF_Q" },
	{ AST_CONTROL_INCOMPLETE, "CONTROL_INCOMPLETE" },
	{ AST_CONTROL_MCID, "CONTROL_MCID" },
	{ AST_CONTROL_UPDATE_RTP_PEER, "CONTROL_UPDATE_RTP_PEER" },
};

#define TRIGGERGOSUBMAXLENGTH   256
struct frame_trigger_data {
	int list_type; /* 0 = white, 1 = black */
	int values[ARRAY_LEN(controltype2str)];
	char triggergosub[TRIGGERGOSUBMAXLENGTH];
};

static void trigger_gosub_action(struct ast_channel *chan, struct ast_frame *frame, 
									enum ast_framehook_event event, struct frame_trigger_data *framedata, const char* typeName);
static struct ast_str* sintesizeExtension(struct ast_str **buf, char* requiredExtension, const char* typeName);

static void datastore_destroy_cb(void *data) {
	ast_free(data);
}

static const struct ast_datastore_info frame_trigger_datastore = {
	.type = "frametrigger",
	.destroy = datastore_destroy_cb
};

static void hook_destroy_cb(void *framedata)
{
	ast_free(framedata);
}

static struct ast_frame *hook_event_cb(struct ast_channel *chan, struct ast_frame *frame, enum ast_framehook_event event, void *data)
{
	int i;
	int valid_frame = 0;
	const char *controlTypeName;		
	struct frame_trigger_data *framedata = data;
	
	if (!frame) {
		return frame;
	}

	if ((event != AST_FRAMEHOOK_EVENT_WRITE) && (event != AST_FRAMEHOOK_EVENT_READ)) {
		return frame;
	}

	if (frame->frametype != AST_FRAME_CONTROL) {
		return frame;
	}

	
	for (i = 0; i < ARRAY_LEN(controltype2str); i++) {
		if (frame->subclass.integer == controltype2str[i].type) {
			controlTypeName = controltype2str[i].str;
			if ((framedata->list_type == 0) && (framedata->values[i])) { /* white list */
				valid_frame = 1;
			} else if ((framedata->list_type == 1) && (!framedata->values[i])){ /* black list */
				valid_frame = 1;
			}
			break;
		}
	}

	if (valid_frame) {
		ast_debug(1, "%s on Channel %s\n", event == AST_FRAMEHOOK_EVENT_READ ? "<--Read" : "--> Write", chan->name);
		trigger_gosub_action(chan, frame, event, framedata, controlTypeName);
	}
	return frame;
}

static int frame_trigger_helper(struct ast_channel *chan, const char *cmd, char *data, const char *value)
{
	struct frame_trigger_data *framedata;
	struct ast_datastore *datastore = NULL;
	struct ast_framehook_interface interface = {
		.version = AST_FRAMEHOOK_INTERFACE_VERSION,
		.event_cb = hook_event_cb,
		.destroy_cb = hook_destroy_cb,
	};
	int i = 0;

	if (!(framedata = ast_calloc(1, sizeof(*framedata)))) {	
		return 0;
	}
	ast_copy_string(framedata->triggergosub, "", TRIGGERGOSUBMAXLENGTH);	

	interface.data = framedata;

	if (strcasestr(data, "black")) {
		framedata->list_type = 1;
	}
	for (i = 0; i < ARRAY_LEN(controltype2str); i++) {
		if (strcasestr(data, controltype2str[i].str)) {
			framedata->values[i] = 1;
		}
	}

	if (value)
		ast_copy_string(framedata->triggergosub, value, TRIGGERGOSUBMAXLENGTH);

	ast_channel_lock(chan);
	i = ast_framehook_attach(chan, &interface);
	if (i >= 0) {
		int *id;
		if ((datastore = ast_channel_datastore_find(chan, &frame_trigger_datastore, NULL))) {
			id = datastore->data;
			ast_framehook_detach(chan, *id);
			ast_channel_datastore_remove(chan, datastore);
		}

		if (!(datastore = ast_datastore_alloc(&frame_trigger_datastore, NULL))) {
			ast_framehook_detach(chan, i);
			ast_channel_unlock(chan);
			return 0;
		}

		if (!(id = ast_calloc(1, sizeof(int)))) {
			ast_datastore_free(datastore);
			ast_framehook_detach(chan, i);
			ast_channel_unlock(chan);
			return 0;
		}

		*id = i; /* Store off the id. The channel is still locked so it is safe to access this ptr. */
		datastore->data = id;
		ast_channel_datastore_add(chan, datastore);
	}
	ast_channel_unlock(chan);

	return 0;
}

static struct ast_custom_function frame_trigger_function = {
	.name = "FRAME_TRIGGER",
	.write = frame_trigger_helper,
};

static int unload_module(void)
{
	return ast_custom_function_unregister(&frame_trigger_function);
}

static int load_module(void)
{
	int res = ast_custom_function_register(&frame_trigger_function);
	return res ? AST_MODULE_LOAD_DECLINE : AST_MODULE_LOAD_SUCCESS;
}

static void trigger_gosub_action(struct ast_channel *chan, struct ast_frame *frame, enum ast_framehook_event event, struct frame_trigger_data *framedata, const char* typeName) 
{
	int res = 0;
	struct ast_app *application;	
	
	if (!framedata || !chan || !frame || !typeName)
		return;
	
	ast_debug(1, "func_frame_trigger: gosub=%s, controlType=%s.\n", framedata->triggergosub, typeName);
	
	if (ast_strlen_zero(framedata->triggergosub))
		return;
	
	application = pbx_findapp("Gosub");
	
	if (application) {
		char *gosub_args, *gosub_argstart;
		char *triggergosub;
		struct ast_str *extension = 0;
		
		triggergosub = ast_strdupa(framedata->triggergosub);
		
		gosub_argstart = strchr(triggergosub, ',');
		if (gosub_argstart) {
			
			*gosub_argstart = 0;
			extension = sintesizeExtension(&extension, gosub_argstart + 1, typeName);
			
			if (!ast_exists_extension(chan, triggergosub, ast_str_buffer(extension), 1, S_COR(chan->caller.id.number.valid, chan->caller.id.number.str, NULL))) {
				extension = sintesizeExtension(&extension, "s", typeName);
			}
		}  else {
			
			extension = sintesizeExtension(&extension, "s", typeName);
		}
		
		if (!ast_exists_extension(chan, triggergosub, ast_str_buffer(extension), 1, S_COR(chan->caller.id.number.valid, chan->caller.id.number.str, NULL))) {
			
			extension = sintesizeExtension(&extension, "s", "");
		}
		
		if (!ast_exists_extension(chan, triggergosub, ast_str_buffer(extension), 1, S_COR(chan->caller.id.number.valid, chan->caller.id.number.str, NULL)) &&
			 ast_exists_extension(chan, triggergosub, "~~s~~", 1, S_COR(chan->caller.id.number.valid, chan->caller.id.number.str, NULL))) {
			extension = sintesizeExtension(&extension, "~~s~~", "");
		}
		
		if (asprintf(&gosub_args, "%s,%s,1", triggergosub, ast_str_buffer(extension)) < 0) {
			ast_log(LOG_WARNING, "asprintf() failed: %s\n", strerror(errno));
			gosub_args = NULL;
		}
		
		if (extension)
			ast_free(extension);
		
		if (gosub_args) {
			res = pbx_exec(chan, application, gosub_args);
			if (!res) {
				struct ast_pbx_args args;
				memset(&args, 0, sizeof(args));
				args.no_hangup_chan = 1;
				ast_pbx_run_args(chan, &args);
			}
			ast_free(gosub_args);
			ast_debug(1, "Gosub exited with status %d\n", res);
		} else {
			ast_log(LOG_ERROR, "Could not Allocate string for Gosub arguments -- Gosub Call Aborted!\n");
		}

	}
}

static struct ast_str* sintesizeExtension(struct ast_str **buf, char* requiredExtension, const char* typeName)
{
	struct ast_str *tmp = *buf;
	
	if (tmp) {
		ast_free(tmp);
	}
	
	tmp = ast_str_create(TRIGGERGOSUBMAXLENGTH);
	ast_str_set(&tmp, 0, "%s", requiredExtension);
	
	if (!ast_strlen_zero(typeName)) {
		ast_str_append(&tmp, 0, "-");
		ast_str_append(&tmp, 0, "%s", typeName);
	}
	
	return tmp;
}

AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "Frame Trace for internal ast_frame debugging.");

